#include "../../common/oldbuffer.h"
#include "../../common/Color.h"
#include "../shaders.h"
#include "../../parse/parser.h"
#include "../../frontend/globals.h"
#include "../../misc/save_frame.h"
#include "../../misc/configDefines.h"
#include <stdlib.h>
#ifdef USE_OMP
#include <omp.h>
#endif

extern Color black;
extern int frameNumber;

#define passes 1
#define tableSize reprojSamples*passes*100*2

#define numFrames 2

typedef struct {
	Camera cameras[numFrames];
	Buffer<Color> *colors[numFrames];
	Buffer<Color> *output[numFrames];
	Buffer<HitPoint> *hits[numFrames];
	Buffer<unsigned int> *accum[numFrames];
	float scaleTable[tableSize];
} reproj_blur_data;

int _init_reproj_blur(void **data_ptr, char *input)
{
	reproj_blur_data *rpd = (reproj_blur_data*) malloc(sizeof(reproj_blur_data));
	
	for(int i=0; i<numFrames; i++)
	{
		rpd->colors[i] = new Buffer<Color>(main_scene->width, main_scene->height);
		rpd->output[i] = new Buffer<Color>(main_scene->width, main_scene->height);
		rpd->hits[i] = new Buffer<HitPoint>(main_scene->width, main_scene->height);		
		rpd->accum[i] = new Buffer<unsigned int>(main_scene->width, main_scene->height);		
		rpd->colors[i]->copy(*main_scene->colorOutput);
		rpd->hits[i]->copy(*main_scene->hitData);
		rpd->cameras[i] = *main_scene->camera;
	}
	
	rpd->scaleTable[0] = 1.0f;
	for(int i=1; i<tableSize; i++)
		rpd->scaleTable[i] = 1.0f/(i+1);

	*data_ptr = (void*)rpd;
	return 1;
}

inline char validReprojectCoords(vector3& newCoords, vector3 &hitPoint, Camera& c)
{
	vector3 newProjection = hitPoint - c.pos;
	c.getImageCoords(newCoords, newProjection);
	
	//if(oldIntersection->normal.dot(newProjection) > 0)
	//	return false;
	if(main_scene->camera->look.dot(newProjection) < 0)
		return false;
	newCoords[0] = roundf(newCoords[0]);
	newCoords[1] = roundf(newCoords[1]);
	if(newCoords.c[0] < 0 || newCoords.c[0] > main_scene->width ||
	   newCoords.c[1] < 0 || newCoords.c[1] > main_scene->height)
		return false;
	
	return true;
}

inline void reprojectBlur(Camera &camFrom, Camera &camTo, 
						  Buffer<Color> *fromBuffer, Buffer<Color> *toBuffer,
						  Buffer<HitPoint> *fromHits, Buffer<unsigned int> *accumulator,
						  int pass)
{
	HitPoint h;
	Color c;
	Camera reprojCamera = camFrom;
	float numSamples = reprojSamples;
	float colorScale = 1.5f/numSamples;
	float bufferScale = 1.0f - colorScale;
	
	if(camFrom.pos == camTo.pos && camFrom.look == camTo.look)
		return;
	
	for(int i=0; i<reprojSamples; i++)
	{
		vector3 newCoords;
		vector3 jitter;
		const float fromAmount = i/numSamples;
		const float toAmount = (1.0f-fromAmount);
		
		const float jitterAmount = (pass+1) * 0.000;
		jitter.randomizeFast();
		jitter *= jitterAmount;
		reprojCamera.pos = camFrom.pos*toAmount + camTo.pos*fromAmount + jitter;
		reprojCamera.look = camFrom.look*toAmount + camTo.look*fromAmount + jitter;
		reprojCamera.horz = camFrom.horz*toAmount + camTo.horz*fromAmount + jitter;
		reprojCamera.vert = camFrom.vert*toAmount + camTo.vert*fromAmount + jitter;
		reprojCamera.topLeftCorner = camFrom.topLeftCorner*toAmount + camTo.topLeftCorner*fromAmount + jitter;
		
		for(int y=0; y<main_scene->height; y++)
		{
			for(int x=0; x<main_scene->width; x++)
			{				
				h = fromHits->getIndex(x,y);			
				if(validReprojectCoords(newCoords, h.pos, reprojCamera))
				{
					c = fromBuffer->getIndex(x,y);
					accumulator->getIndex(newCoords[0], newCoords[1])++;
					toBuffer->getIndex(newCoords[0], newCoords[1]) += c;
				}
			}
		}
	}
}

void _run_reproj_blur(void *data)
{
	reproj_blur_data *rpd = (reproj_blur_data*)data;
	
	for(int i=0; i<numFrames; i++)
		rpd->accum[i]->zeroData();
	
	//shift color buffers and get the new frame
	rpd->colors[0]->swapData(*rpd->colors[1]);
	rpd->colors[1]->copy(*main_scene->colorOutput);
	
	//shift camera data and get new camera
	rpd->cameras[0] = rpd->cameras[1];
	rpd->cameras[1] = *main_scene->camera;
	
	//shift hit data and get new data
	rpd->hits[0]->swapData(*rpd->hits[1]);
	rpd->hits[1]->swapData(*main_scene->hitData);

	rpd->output[0]->reset();
	rpd->output[1]->reset();
	
	//#pragma omp parallel for schedule(static, 1)
	rpd->output[0]->copy(*main_scene->colorOutput);
	
	//reproject from old camera and to new camera
	//omp_set_num_threads(2);
	#pragma omp parallel for schedule(static, 1)
	for(int i=0; i<numFrames; i++)
		reprojectBlur(rpd->cameras[0], rpd->cameras[1], rpd->colors[i], rpd->output[i], rpd->hits[i], rpd->accum[i], 0);
	
	//combine frames
	#pragma omp parallel for schedule(guided, 100)
	for(int i=0; i<rpd->accum[1]->getCount(); i++)
	{
		//rpd->output[0]->getIndex(i) += rpd->output[1]->getIndex(i);
		//rpd->accum[0]->getIndex(i) += rpd->accum[1]->getIndex(i);
		main_scene->colorOutput->getIndex(i) = rpd->output[0]->getIndex(i) * rpd->scaleTable[ rpd->accum[0]->getIndex(i) ];
	}
}


void _cleanup_reproj_blur(void* data)
{
	free((reproj_blur_data*)data);
}

shader_data reproj_blur =
{
	"reproj_blur",
	"options:\n   reproj_blur <float alpha>\n",
	(int (*)(void*, char*))_init_reproj_blur,
	(void* (*)(void*))_run_reproj_blur,
	(void* (*)(void*))_cleanup_reproj_blur,
};

shader_data* get_reproj_blur()
{
	return &reproj_blur;
}
